#include <stdio.h>
#include <math.h>
#include "erlang.h"

/************************************************************

The Qlib program library has resulted from the work conducted
over several years jointly by

 Laboratory of Telecommunications Technology of
 Helsinki University of Technology (HUT) and

 VTT Information Technology, Telecommunications

The following individuals have contributed to the library:

 Samuli Aalto, Jani Lakkakorpi, Ilkka Norros, Anssi Pirhonen
 and Jorma Virtamo.

The programs can be used and modified freely.

No claims are made about the correctness of the programs
and no liability is taken for any damage caused by the use
of the programs in the library.

Reports of bugs in the program can be sent to qlib@tct.hut.fi

*************************************************************/

static double cont_frac_erl(double, double);

/*
 * Erlang's blocking function, the number of lines is an integer.
 * int n    = The number of lines 
 * double a = Traffic intensity
 */

double
Berl_i (int n, double a)
{
  int i;
  double e = 1;
  
  for (i = 1; i <= n; i++) {
    e = a*e/(i+a*e);
  }
  return(e);
}

/*
 * Erlang's blocking function, the number of lines is a real.
 * double x = The number of lines
 * double a = Traffic intensity
 *
 */

double
Berl_d (double x, double a)
{
  double zx = x - floor(x);
  double b,s,z,u,term,zold,znew,xp1;
  int k;
  
  if (x<=0) return 1;
  
  
  
  if (x < a && a >0.01) { 
    /* Method 1, continued fraction expansion */
    
    return(cont_frac_erl(x, a));
  }
  
  if (x > 60) {
    /* Method 2 */
    
    k = floor(0.9*a);
    z = k + zx;
    b = Berl_d(z,a);
    s = 1/a;
    u = x/a+s/2;
    for (z = (z+1)/a; z < u; z +=s) {
      b = b/(z+b);
      if (b == 0) return b;
    }
    return b;
  }
  
  if (x >= 1) {
    /* Method 3 */
    
    b = (zx/(a+1)-a-2) / ((a+3)*a+1)*zx + 1;
    u = (x + 0.5) / a;
    s = 1/a;
    for (zx = (zx+1)/a; zx < u; zx += s)
      b = b/(zx+b);
    return(b);
  }
  
  if (a >0.01) {
    /* Method 4 */
    
    b = 0;
    b +=  pow(1+38.530683306486/a, x) * 1.48302705111 * 1e-16;
    b +=  pow(1+31.407519169754/a, x) * 1.45651526407 * 1e-13; 
    b +=  pow(1+25.623894226729/a, x) * 3.92189726704 * 1e-11; 
    b +=  pow(1+20.776478899449/a, x) * 4.22743038498 * 1e-9; 
    b +=  pow(1+16.654407708330/a, x) * 2.22631690710 * 1e-7;
    b +=  pow(1+13.130282482176/a, x) * 6.45992676202 * 1e-6;
    b +=  pow(1+10.120228568019/a, x) * 1.11674392344 * 1e-4;
    b +=  pow(1+ 7.565916226613/a, x) * 1.21243614721 * 1e-3;
    b +=  pow(1+ 5.425336627414/a, x) * 8.56387780361 * 1e-3;
    b +=  pow(1+ 3.667622721751/a, x) * 4.02068649210 * 1e-2;
    b +=  pow(1+ 2.269949526204/a, x) * 1.26425818106 * 1e-1;
    b +=  pow(1+ 1.215595412071/a, x) * 2.63027577942 * 1e-1;
    b +=  pow(1+ 0.492691740302/a, x) * 3.42210177923 * 1e-1;
    b +=  pow(1+ 0.093307812017/a, x) * 2.18234885940 * 1e-1;
    
    return 1/b;
  }
  
  
  /* Method 5 */
  
  xp1 = x + 1;
  
  znew = zold = 
    sqrt(2*M_PI) * sqrt(xp1) / 
    exp( x * log(a/xp1) - a + x + 1 -  1/xp1/12 + 1 / (360 * xp1 * xp1 * xp1));
  term = 1;
  
  
  while (fabs(term/zold*znew) > 0.0001) {
    x = x + 1;     
    zold = znew;
    term = term * a/x;
    znew = zold-term;
  }
  return(1/znew);
  
}



#define MAX_ITERATIONS 200

static double
cont_frac_erl(double x, double A)
{
  double a,b;
  double An_2, An_1, An;
  double Bn_2, Bn_1, Bn;
  int n;
  
  An_2 = 1;
  An_1 = 0;
  Bn_2 = 0;
  Bn_1 = 1;
  a = A;
  b = A;
  
  for (n = 1; n <= MAX_ITERATIONS; n++) {
    An = b*An_1 + a*An_2;
    Bn = b*Bn_1 + a*Bn_2;
    if ((Bn_1 == Bn &&  An_1 == An) || An > 1e50 || Bn > 1e50) {
      break;
    }
    
    b = (b == A) ? 1 : A;
    a =  n /2;
    if (n % 2) 
      a -= x;
    An_2 = An_1;
    Bn_2 = Bn_1;
    An_1 = An;
    Bn_1 = Bn;
  }
  
  return(Bn/An);
}

